import os
import uuid
import json
from mitoinstaller import __version__
from datetime import datetime

# Where all global .mito files are stored
MITO_FOLDER = os.path.join(os.path.expanduser("~"), '.mito')

# The path of the user.json file, which notably is the same
# path as the USER_JSON_PATH in mitosheet
USER_JSON_PATH = os.path.join(MITO_FOLDER, 'user.json')

def get_random_id():
    """
    Creates a new random ID for the user, which for any given user,
    should only happen once.
    """
    return str(uuid.uuid1())

def is_running_test():
    """
    A helper function that quickly returns if the current code is running inside 
    of a test, which is useful for making sure we don't generate tons of logs 
    """
    running_pytests = "PYTEST_CURRENT_TEST" in os.environ
    running_ci = 'CI' in os.environ and os.environ['CI'] is not None

    return running_pytests or running_ci


# NOTE: the installer only creates the static id for the user, and
# otherwise does nothing with the user_json file. This makes sure
# we keep the dependencies as simple as possible with this file
USER_JSON_DEFAULT = {
    'static_user_id': get_random_id() if not is_running_test() else 'github_action',
}

def try_create_user_json_file():
    # Create the mito folder if it does not exist
    if not os.path.exists(MITO_FOLDER):
        os.mkdir(MITO_FOLDER)
        
    # We only create a user.json file if it does not exist
    if not os.path.exists(USER_JSON_PATH):
        with open(USER_JSON_PATH, 'w+') as f:
            # And write the default object
            f.write(json.dumps(USER_JSON_DEFAULT))
    
def get_static_user_id():
    try:
        with open(USER_JSON_PATH) as f:
            return json.load(f)['static_user_id']
    except: 
        return None

def user_json_only_has_static_user_id():
    """
    Returns True if the user.json file only has the static_user_id in it
    and otherwise returns False
    """
    try:
        with open(USER_JSON_PATH) as f:
            user_json_object = json.load(f)
            return len(user_json_object) == 1 and 'static_user_id' in user_json_object
    except:
        return False


"""
Useful functions and utilities for
logging information about install
"""
import traceback
import analytics
import platform
import getpass

analytics.write_key = '6I7ptc5wcIGC4WZ0N1t0NXvvAbjRGUgX' 

from mitoinstaller import __version__
from mitoinstaller.user_install import get_static_user_id, user_json_only_has_static_user_id

def is_on_kuberentes_mito():
    """
    Returns True if the user is on Kuberentes Mito, on staging or on app
    """
    user = getpass.getuser()
    return user == 'jovyan'

def is_local_deployment():
    """
    Helper function for figuring out if this a local deployment or a
    Mito server deployment
    """
    return not is_on_kuberentes_mito() 

def identify():
    """
    This identify call identifies the user in our logs. Note that
    we make sure that the actual identify call only runs when
    the user has a user.json with only a static id, as if we
    have more feilds, then the user has already run the mitosheet.sheet
    call and thus been identified there.
    """
    static_user_id = get_static_user_id()
    operating_system = platform.system()

    if user_json_only_has_static_user_id():
        analytics.identify(
            static_user_id,
            {
                'operating_system': operating_system,
                'version_installer': __version__,
                'local': is_local_deployment()
            }
        )

def log_step_failed(install_or_upgrade, package, reason):
    # First, we print the traceback, to make live debugging easier
    print(traceback.format_exc())
    # Then, we log it
    log(
        '{install_or_upgrade}_failed_{package}'.format(install_or_upgrade=install_or_upgrade, package=package),
        {  
            'package': package,
            'reason': reason, 
            'error_traceback': traceback.format_exc().split('\n')
        }
    )

def log(event, params=None):
    """
    A utility that all logging should pass through
    """
    static_user_id = get_static_user_id()

    if params is None:
        params = {}

    analytics.track(
        static_user_id, 
        event, 
        params
    )



"""
The Mito Installer package contains utils for installing
Mito within your Python enviornment.

Long term, we aim to meet:
1. This package has minimal dependencies, both for speed of download and the ultimate portability.
2. The installation attempts to fail as early as possible, and to give the user as much help
   help as possible while doing so.
"""
from mitoinstaller.install import do_install_or_upgrade

def main():
    """
    The main function of the Mito installer, this function is responsible
    for either installing mitosheet or upgrading mitosheet.

    To install Mito (for the first time):
    python -m mitoinstaller install

    To upgrade Mito:
    python -m mitoinstaller upgrade
    """
    import sys

    if len(sys.argv) > 1:
        command = sys.argv[1]
    else:
        command = ''

    if command == 'install':
        do_install_or_upgrade('install')
    elif command == 'uninstall':
        print('To uninstall, run `pip uninstall mitosheet` and `pip uninstall mitosheet3`')
    elif command == 'upgrade':
        do_install_or_upgrade('upgrade')
    else:
        print('\nProper usage is `python -m mitoinstaller install` or `python -m mitoinstaller upgrade`\n\nTry running the command `python -m mitoinstaller install`\n')
 

if __name__ == '__main__':
    main()


"""
We specify the installer as a list of steps that 
must run in order, for the installer.

Currently, we take the following steps:
1. First, try and install mitosheet3. Because mitosheet3 uses 
   JLab 3, we first check if the user has any conflicting
   JLab 2 dependencies installed. If they do, we abort. If they 
   don't, we update them to JLab 3.
2. Then, if the above installation of mitosheet3 fails for 
   any reason, we try to install mitosheet on JLab 2. 
3. If neither of them work, we give up, sadly. 
"""

from mitoinstaller.commands import check_running_jlab_3_processes, install_pip_packages, jupyter_labextension_list, uninstall_labextension, uninstall_pip_packages, upgrade_mito_installer, exit_with_error
from mitoinstaller.log_utils import log, identify, log_step_failed
from mitoinstaller.user_install import (try_create_user_json_file, get_static_user_id)
from mitoinstaller import __version__

# SHARED UTILITIES

def initial_install_step_create_user(install_or_upgrade):
    static_user_id = get_static_user_id()

    # If the user has no static install ID, create one
    if static_user_id is None:
        try_create_user_json_file()    
    
    identify()
    log('{install_or_upgrade}_started'.format(install_or_upgrade=install_or_upgrade), {
        'mitoinstaller_version': __version__
    })

def get_extension_names_from_labextension_list_output(stdout, stderr):
    """
    Returns a list of the extension names installed from the stdout, stderr
    pair returned from the jupyter labextension list.

    If you run the command: `jupyter labextension list`, the output is: 
    ------
    JupyterLab v0.35.0
    Known labextensions:
    app dir: /Users/nate/saga-vcs/monorepo/mito/installer/venv/share/jupyter/lab
            @jupyter-widgets/jupyterlab-manager v0.38.1 enabled OK
    ------

    Note that part of this output prints to stdout, and other parts to stderr 
    (for some reason), so we append them with a newline so we make sure we 
    get all of the extensions correctly!
    """
    def is_extension_line(line):
        # Check that it has a version
        if len(line) == 0:
            return False

        if 'v' not in line:
            return False
        
        # Check that it is either enabled or disabled
        return 'enabled' in line or 'disabled' in line

    output = stdout + "\n" + stderr
    extension_lines = [line.strip() for line in output.splitlines() if is_extension_line(line)]
    extension_names = []
    for line in extension_lines:
        extension_names.append(line.split(" ")[0])

    return extension_names

def get_jupyterlab_metadata():
    """
    Helper function that returns a tuple of: (jupyterlab_version), (installed_extensions)

    If no JupyterLab is installed, returns (None, None)
    """
    try:
        from jupyterlab import __version__
    except Exception as e:
        # If this import fails, it must not be installed
        return None, None

    stdout, stderr = jupyter_labextension_list()
    extension_names = get_extension_names_from_labextension_list_output(stdout, stderr)
    return __version__, extension_names


def install_step_upgrade_mitoinstaller():
    upgrade_mito_installer()
    


# MITOSHEET 3 INSTALL STEPS

def install_step_mitosheet3_check_dependencies():
    """
    This is the most complex step in the installation process. It's
    goal is to check if the users existing installation can safely
    be upgraded to JLab 3.0. 

    To do this, it checks a variety of conditions, mostly around what
    version of JLab they have installed, and if this version of JLab has
    any installed dependencies (that we cannot safely upgrade).
    """

    jupyterlab_version, extension_names = get_jupyterlab_metadata()

    # If no JupyterLab is installed, we can continue with install, as
    # there are no conflict dependencies
    if jupyterlab_version is None:
        return
    
    # If JupyterLab 3 is installed, then we are are also good to go
    if jupyterlab_version.startswith('3'):
        return
    
    if len(extension_names) == 0:
        return
    elif len(extension_names) == 1 and extension_names[0] == 'mitosheet':
        uninstall_labextension('mitosheet')
        uninstall_pip_packages('mitosheet')
        log('uninstalled_mitosheet_labextension',
            {
                'jupyterlab_version': jupyterlab_version,
                'extension_names': extension_names
            }
        )
        return
    
    else:
        raise Exception('Installed extensions {extension_names}'.format(extension_names=extension_names))


def install_step_mitosheet3_install_mitosheet3():
    install_pip_packages('mitosheet3')


INSTALL_MITOSHEET3_STEPS = [
    {
        'step_name': 'Upgrading mitoinstaller',
        'execute': install_step_upgrade_mitoinstaller
    },
    {
        'step_name': 'Checking dependencies',
        'execute': install_step_mitosheet3_check_dependencies
    },
    {
        'step_name': 'Installing mitosheet3',
        'execute': install_step_mitosheet3_install_mitosheet3
    },
]


# MITOSHEET 2 INSTALL STEPS


def install_step_mitosheet_check_dependencies():
    jupyterlab_version, extension_names = get_jupyterlab_metadata()

    # If no JupyterLab is installed, we can continue with install, as
    # there are no conflict dependencies
    if jupyterlab_version is None:
        return
    
    # If JupyterLab 2 is installed, then we are are also good to go
    if jupyterlab_version.startswith('2'):
        return
    
    if len(extension_names) == 0:
        return
    elif len(extension_names) == 1 and extension_names[0] == 'mitosheet':
        return
    else:
        raise Exception('Installed extensions {extension_names}'.format(extension_names=extension_names))


def install_step_mitosheet_install_mitosheet():
    install_pip_packages('mitosheet')


def install_step_mitosheet_install_jupyter_widget_manager():
    from jupyterlab import commands
    commands.install_extension('@jupyter-widgets/jupyterlab-manager@2')


def install_step_mitosheet_rebuild_jupyterlab():
    from jupyterlab import commands
    commands.build()


INSTALL_MITOSHEET_STEPS = [
    {
        'step_name': 'Upgrading mitoinstaller',
        'execute': install_step_upgrade_mitoinstaller
    },
    {
        'step_name': 'Checking dependencies',
        'execute': install_step_mitosheet_check_dependencies
    },
    {
        'step_name': 'Installing mitosheet',
        'execute': install_step_mitosheet_install_mitosheet
    },
    {
        'step_name': 'Installing @jupyter-widgets/jupyterlab-manager@2',
        'execute': install_step_mitosheet_install_jupyter_widget_manager
    },
    {
        'step_name': 'Rebuilding JupyterLab',
        'execute': install_step_mitosheet_rebuild_jupyterlab
    },
]


def get_success_message(install_or_upgrade):
    """
    We show a different message depending on if this is an install or an upgrade,
    and we further tell the user different things if they have a currently running
    JLab instance (as they need to restart this)
    """
    running_jlab = check_running_jlab_3_processes()

    separator_line = '----------------------------------------------------------------------------'

    install_start = 'Mito has finished installing'
    upgrade_start = 'Mito has finished upgrading.'

    launch_jlab = 'Launch JupyterLab with:\tpython -m jupyter lab'
    relaunch_jlab = 'Please shut down the currently running JupyterLab and relaunch it to enable Mito'

    render_mitosheet = 'Then render a mitosheet following the instructions here: https://docs.trymito.io/how-to/creating-a-mitosheet'

    if not running_jlab:
        if install_or_upgrade == 'install':
            return '\n{separator_line}\n{install_start}\n\n{launch_jlab}\n\n{render_mitosheet}\n{separator_line}'.format(
                separator_line=separator_line,
                install_start=install_start,
                launch_jlab=launch_jlab,
                render_mitosheet=render_mitosheet,
            )
        else:
            return '\n{separator_line}\n{upgrade_start}\n\n{launch_jlab}\n\n{render_mitosheet}\n{separator_line}'.format(
                separator_line=separator_line,
                upgrade_start=upgrade_start,
                launch_jlab=launch_jlab,
                render_mitosheet=render_mitosheet,
            )
    else:
        if install_or_upgrade == 'install':
            return '\n{separator_line}\n{install_start}\n\n{relaunch_jlab}\n\n{render_mitosheet}\n{separator_line}'.format(
                separator_line=separator_line,
                install_start=install_start,
                relaunch_jlab=relaunch_jlab,
                render_mitosheet=render_mitosheet,
            )
        else:
            return '\n{separator_line}\n{upgrade_start}\n\n{relaunch_jlab}\n\n{render_mitosheet}\n{separator_line}'.format(
                separator_line=separator_line,
                upgrade_start=upgrade_start,
                relaunch_jlab=relaunch_jlab,
                render_mitosheet=render_mitosheet,
            )

def do_install_or_upgrade(install_or_upgrade):
    """
    install_or_upgrade is the workhorse of actually installing mito. It first attempts
    to install mitosheet3, and if that fails, installs mitosheet.

    install_or_upgrade should be 'install' or 'upgrade'

    Notably, the process for installing Mito initially and upgrading Mito are
    identical. As such, we reuse this function to upgrade, just with different
    error and logging messages.
    """
    print("Starting {install_or_upgrade}...".format(install_or_upgrade=install_or_upgrade))

    initial_install_step_create_user(install_or_upgrade)

    for idx, install_step in enumerate(INSTALL_MITOSHEET3_STEPS):
        try:
            # Print the install step name
            print(install_step['step_name'])
            # Actually execute the step
            install_step['execute']()
            # If we finish install, then tell the user, and exit
            if idx >= len(INSTALL_MITOSHEET3_STEPS) - 1:
                log('{install_or_upgrade}_finished_mitosheet3'.format(install_or_upgrade=install_or_upgrade), {'package': 'mitosheet3'})
                print(get_success_message(install_or_upgrade))
                exit(0)
        except SystemExit:
            exit(0)
        except:
            # Note: if mitosheet3 installation fails, we log it, but we do
            # not exit, as we want to continue and try to install mitosheet
            error_message = "Failed: " + install_step['step_name']
            log_step_failed(install_or_upgrade, 'mitosheet3', error_message)
            print(error_message)
            # We do break out of this loop!
            break

    for install_step in INSTALL_MITOSHEET_STEPS:
        try:
            print(install_step['step_name'])
            install_step['execute']()
        except:
            error_message = "Failed: " + install_step['step_name']
            log_step_failed(install_or_upgrade, 'mitosheet', error_message)
            exit_with_error(install_or_upgrade, error_message)

    log('{install_or_upgrade}_finished_mitosheet'.format(install_or_upgrade=install_or_upgrade), {'package': 'mitosheet'})
    print(get_success_message(install_or_upgrade))


"""
Contains useful commands for interacting
with the command line directly
"""

# NOTE: Do not import subprocess here, we only want one
# function to have it
import sys

def get_output(completed_process):
    output = ""
    if completed_process.stdout is not None:
        output += completed_process.stdout
    if completed_process.stderr is not None:
        output += completed_process.stderr
    return output


def run_command(command_array):
    """
    An internal command that should be used to run all commands
    that run on the command line, so that output from failing
    commands can be captured.
    """
    import subprocess
    completed_process = subprocess.run(
        command_array, 
        # NOTE: we do not use the capture_output variable, as this doesn't work before
        # python 3.7, and we want users to be able to install before that
        stdout=subprocess.PIPE, 
        stderr=subprocess.STDOUT,
        # NOTE: we use universal_newlines to get the result back as text, 
        # but we don't use text=True, because we want to work before 3.7 when
        # text was introduced. See here: https://stackoverflow.com/questions/41171791/how-to-suppress-or-capture-the-output-of-subprocess-run
        universal_newlines=True
    )
    if completed_process.returncode != 0:
        raise Exception(get_output(completed_process))

    # We default the stdout and stderr to empty strings if they are not strings, 
    # to make code that handles them have an easier time (they might be None)
    stdout = completed_process.stdout if isinstance(completed_process.stdout, str) else ''
    stderr = completed_process.stderr if isinstance(completed_process.stderr, str) else ''
    return stdout, stderr

def jupyter_labextension_list():
    """
    Returns the stdout, stderr pair for the currently
    installed jupyterlab extensions.
    """

    sys_call = [sys.executable, "-m", "jupyter", "labextension", "list"]
    return run_command(sys_call)


def uninstall_labextension(extension):
    """
    Uninstall a labextension
    """

    sys_call = [sys.executable, "-m", "jupyter", "labextension", "uninstall", extension]
    run_command(sys_call)


def uninstall_pip_packages(*packages):
    """
    This function uninstalls the given packages in a single pass
    using pip, through the command line.
    """

    sys_call = [sys.executable, "-m", "pip", "uninstall", "-y"]

    for package in packages:
        sys_call.append(package)
    
    run_command(sys_call)


def install_pip_packages(*packages):
    """
    This function installs the given packages in a single pass
    using pip, through the command line.

    https://stackoverflow.com/questions/12332975/installing-python-module-within-code
    """

    sys_call = [sys.executable, "-m", "pip", "install"]

    for package in packages:
        sys_call.append(package)
    sys_call.append('--upgrade')

    run_command(sys_call)

def upgrade_mito_installer():
    """
    Upgrades the mito installer package itself
    """
    run_command([sys.executable, "-m", "pip", "install", 'mitoinstaller', '--upgrade', '--no-cache-dir'])


def check_running_jlab_3_processes():
    """
    Returns true if there are running JLab 3 processes, 
    returns false if there are not.

    Useful for telling the user to refresh their servers
    if they install Mito
    """
    sys_call = [sys.executable, "-m", "jupyter", "server", "list"]
    stdout, stderr = run_command(sys_call)
    return len(stdout.strip().splitlines()) > 1

def check_running_jlab_not_3_processes():
    """
    Returns true if there are running JLab processes, 
    returns false if there are not.

    Useful for telling the user to refresh their servers
    if they install Mito
    """
    sys_call = [sys.executable, "-m", "jupyter", "notebook", "list"]
    stdout, stderr = run_command(sys_call)
    return len(stdout.strip().splitlines()) > 1

def check_running_jlab_processes():
    """
    Returns true if there are running JLab processes from any version
    returns false if there are not.
    """
    return check_running_jlab_3_processes() or check_running_jlab_not_3_processes()

def exit_with_error(install_or_upgrade, error=None):
    full_error = '\n\nSorry, looks like we hit a problem during {install_or_upgrade}. '.format(install_or_upgrade=install_or_upgrade) + \
        ('' if error is None else ("It seems we " + error + '.')) + \
        '\nWe\'re happy to help you fix it, just shoot an email to jake@sagacollab.com and copy in the output above.\n'

    print(full_error)
    exit(1)


from distutils.core import setup
from mitoinstaller import __version__

setup(
    name='mitoinstaller',
    version=__version__,
    packages=['mitoinstaller',],
    install_requires=['analytics-python'],
    license='TODO',
    long_description='The mitoinstaller package allows you to install the mitosheet package in your local JupyterLab instance. It logs anonymous data about installation, including if it is successful or why it failed.',
)
